import os
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from phoenix.evals.models.base import BaseModel
from phoenix.evals.models.rate_limiters import RateLimiter
from phoenix.evals.templates import MultimodalPrompt, PromptPartContentType

DEFAULT_MISTRAL_MODEL = "mistral-large-latest"
"""Use the latest large mistral model by default."""

MINIMUM_MISTRAL_VERSION = "1.0.0"


class MistralRateLimitError(Exception):
    pass


@dataclass
class MistralAIModel(BaseModel):
    """
    An interface for using MistralAI models.

    This class wraps the MistralAI SDK for use with Phoenix LLM evaluations. Calls to the
    MistralAI API are dynamically throttled when encountering rate limit errors. Requires the
    `mistralai` package to be installed.

    Supports Async: ✅
        If possible, makes LLM calls concurrently.

    Args:
        model (str, optional): The model name to use. Defaults to "mistral-large-latest".
        temperature (float, optional): Sampling temperature to use. Defaults to 0.0.
        top_p (float, optional): Total probability mass of tokens to consider at each step.
            Defaults to None.
        random_seed (int, optional): Random seed to use for sampling. Defaults to None.
        response_format (Dict[str, str], optional): A dictionary specifying the format of the
            response. Defaults to None.
        safe_prompt (bool, optional): Whether to use safe prompt. Defaults to False.
        initial_rate_limit (int, optional): The initial internal rate limit in allowed requests
            per second for making LLM calls. This limit adjusts dynamically based on rate
            limit errors. Defaults to 5.
        timeout (int, optional): The timeout for completion requests in seconds. Defaults to 120.

    Example:
        .. code-block:: python

            # Get your own Mistral API Key: https://docs.mistral.ai/#api-access
            # Set the MISTRAL_API_KEY environment variable

            from phoenix.evals import MistralAIModel
            model = MistralAIModel(model="mistral-large-latest")
    """

    model: str = DEFAULT_MISTRAL_MODEL
    temperature: float = 0
    top_p: Optional[float] = None
    random_seed: Optional[int] = None
    response_format: Optional[Dict[str, str]] = None
    api_key: Optional[str] = os.getenv("MISTRAL_API_KEY")
    safe_prompt: bool = False
    initial_rate_limit: int = 5
    timeout: int = 120

    def __post_init__(self) -> None:
        self._init_client()
        self._init_rate_limiter()

    @property
    def _model_name(self) -> str:
        return self.model

    def _init_client(self) -> None:
        try:
            from mistralai import Mistral
            from mistralai.models import SDKError
        except ImportError:
            self._raise_import_error(
                package_name="mistralai",
                package_min_version=MINIMUM_MISTRAL_VERSION,
            )

        self._client = Mistral(api_key=self.api_key)
        self._mistral_sdk_error = SDKError

    def _init_rate_limiter(self) -> None:
        self._rate_limiter = RateLimiter(
            rate_limit_error=MistralRateLimitError,
            max_rate_limit_retries=10,
            initial_per_second_request_rate=self.initial_rate_limit,
            enforcement_window_minutes=1,
        )

    def invocation_parameters(self) -> Dict[str, Any]:
        params = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "random_seed": self.random_seed,
            "safe_prompt": self.safe_prompt,
            "response_format": self.response_format,
        }
        # Mistral is strict about not passing None values to the API
        return {k: v for k, v in params.items() if v is not None}

    def _generate(self, prompt: Union[str, MultimodalPrompt], **kwargs: Dict[str, Any]) -> str:
        if isinstance(prompt, str):
            prompt = MultimodalPrompt.from_string(prompt)

        # instruction is an invalid input to Mistral models, it is passed in by
        # BaseEvalModel.__call__ and needs to be removed
        kwargs.pop("instruction", None)
        invocation_parameters = self.invocation_parameters()
        invocation_parameters.update(kwargs)
        response = self._rate_limited_completion(
            model=self.model,
            messages=self._format_prompt(prompt),
            **invocation_parameters,
        )

        return str(response)

    def _rate_limited_completion(self, **kwargs: Any) -> Any:
        @self._rate_limiter.limit
        def _completion(**kwargs: Any) -> Any:
            try:
                response = self._client.chat.complete(**kwargs)
            # if an SDKError is raised, check that it's a rate limit error:
            except self._mistral_sdk_error as exc:
                http_status = getattr(exc, "http_status", None)
                if http_status and http_status == 429:
                    raise MistralRateLimitError() from exc
                raise exc

            if response and (choices := response.choices):
                if first_response := choices[0]:
                    return first_response.message.content

        return _completion(**kwargs)

    async def _async_generate(
        self, prompt: Union[str, MultimodalPrompt], **kwargs: Dict[str, Any]
    ) -> str:
        # instruction is an invalid input to Mistral models, it is passed in by
        # BaseEvalModel.__call__ and needs to be removed
        if isinstance(prompt, str):
            prompt = MultimodalPrompt.from_string(prompt)

        kwargs.pop("instruction", None)
        invocation_parameters = self.invocation_parameters()
        invocation_parameters.update(kwargs)
        response = await self._async_rate_limited_completion(
            model=self.model,
            messages=self._format_prompt(prompt),
            **invocation_parameters,
        )

        return str(response)

    async def _async_rate_limited_completion(self, **kwargs: Any) -> Any:
        @self._rate_limiter.alimit
        async def _async_completion(**kwargs: Any) -> Any:
            try:
                response = await self._client.chat.complete_async(**kwargs)
            except self._mistral_sdk_error as exc:
                http_status = getattr(exc, "http_status", None)
                if http_status and http_status == 429:
                    raise MistralRateLimitError() from exc
                raise exc

            if response and (choices := response.choices):
                if first_response := choices[0]:
                    return first_response.message.content

        return await _async_completion(**kwargs)

    def _format_prompt(self, prompt: MultimodalPrompt) -> List[Dict[str, str]]:
        messages = []
        for part in prompt.parts:
            if part.content_type == PromptPartContentType.TEXT:
                messages.append({"role": "user", "content": part.content})
            else:
                raise ValueError(
                    f"Unsupported content type for {MistralAIModel.__name__}: {part.content_type}"
                )
        return messages



from __future__ import annotations

import inspect
import logging
import warnings
from collections import defaultdict
from enum import Enum
from functools import wraps
from itertools import product
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import pandas as pd
from pandas import DataFrame
from typing_extensions import TypeAlias

from phoenix.evals.evaluators import LLMEvaluator
from phoenix.evals.exceptions import PhoenixTemplateMappingError
from phoenix.evals.executors import ExecutionStatus, get_executor_on_sync_context
from phoenix.evals.models import BaseModel, OpenAIModel, set_verbosity
from phoenix.evals.templates import (
    ClassificationTemplate,
    MultimodalPrompt,
    PromptOptions,
    PromptTemplate,
    normalize_classification_template,
)
from phoenix.evals.utils import (
    NOT_PARSABLE,
    get_tqdm_progress_bar_formatter,
    openai_function_call_kwargs,
    parse_openai_function_call,
    printif,
    snap_to_rail,
)

logger = logging.getLogger(__name__)


ColumnName: TypeAlias = str
Label: TypeAlias = str
Score: TypeAlias = Optional[float]
Explanation: TypeAlias = Optional[str]
Record: TypeAlias = Mapping[str, Any]
Index: TypeAlias = int

# snapped_response, explanation, response
ParsedLLMResponse: TypeAlias = Tuple[Optional[str], Optional[str], str, str]


class ClassificationStatus(Enum):
    DID_NOT_RUN = ExecutionStatus.DID_NOT_RUN.value
    COMPLETED = ExecutionStatus.COMPLETED.value
    COMPLETED_WITH_RETRIES = ExecutionStatus.COMPLETED_WITH_RETRIES.value
    FAILED = ExecutionStatus.FAILED.value
    MISSING_INPUT = "MISSING INPUT"


PROCESSOR_TYPE = TypeVar("PROCESSOR_TYPE")


def deprecate_dataframe_arg(func: Callable[..., Any]) -> Callable[..., Any]:
    # Remove this once the `dataframe` arg in `llm_classify` is no longer supported

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        signature = inspect.signature(func)

        if "dataframe" in kwargs:
            warnings.warn(
                "`dataframe` argument is deprecated; use `data` instead",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs["data"] = kwargs.pop("dataframe")
        bound_args = signature.bind_partial(*args, **kwargs)
        bound_args.apply_defaults()
        return func(*bound_args.args, **bound_args.kwargs)

    return wrapper


@deprecate_dataframe_arg
def llm_classify(
    data: Union[pd.DataFrame, List[Any]],
    model: BaseModel,
    template: Union[ClassificationTemplate, PromptTemplate, str],
    rails: List[str],
    data_processor: Optional[Callable[[PROCESSOR_TYPE], PROCESSOR_TYPE]] = None,
    system_instruction: Optional[str] = None,
    verbose: bool = False,
    use_function_calling_if_available: bool = True,
    provide_explanation: bool = False,
    include_prompt: bool = False,
    include_response: bool = False,
    include_exceptions: bool = False,
    max_retries: int = 10,
    exit_on_error: bool = True,
    run_sync: bool = False,
    concurrency: Optional[int] = None,
    progress_bar_format: Optional[str] = get_tqdm_progress_bar_formatter("llm_classify"),
) -> pd.DataFrame:
    """
    Classifies each input row of the dataframe using an LLM.
    Returns a pandas.DataFrame where the first column is named `label` and contains
    the classification labels. An optional column named `explanation` is added when
    `provide_explanation=True`.

    Args:
        data (Union[pd.DataFrame, List[Any]): A collection of data which
            can contain template variables and other information necessary to generate evaluations.
            If a passed a DataFrame, there must be column names that match the template variables.
            If passed a list, the elements of the list will be mapped to the template variables
            in the order that the template variables are defined.

        model (BaseEvalModel): An LLM model class.

        template (Union[ClassificationTemplate, PromptTemplate, str]): The prompt template
            as either an instance of PromptTemplate, ClassificationTemplate or a string.
            If a string, the variable names should be surrounded by curly braces so that
            a call to `.format` can be made to substitute variable values.

        rails (List[str]): A list of strings representing the possible output classes
            of the model's predictions.

        data_processor (Optional[Callable[[T], T]]): An optional callable that is used to process
            the input data before it is mapped to the template variables. This callable is passed
            a single element of the input data and can return either a pandas.Series with indices
            corresponding to the template variables or an iterable of values that will be mapped
            to the template variables in the order that the template variables are defined.

        system_instruction (Optional[str], optional): An optional system message.

        verbose (bool, optional): If True, prints detailed info to stdout such as
            model invocation parameters and details about retries and snapping to rails.
            Default False.

        use_function_calling_if_available (bool, default=True): If True, use function
            calling (if available) as a means to constrain the LLM outputs.
            With function calling, the LLM is instructed to provide its response as a
            structured JSON object, which is easier to parse.

        provide_explanation (bool, default=False): If True, provides an explanation
            for each classification label. A column named `explanation` is added to
            the output dataframe.

        include_prompt (bool, default=False): If True, includes a column named `prompt`
            in the output dataframe containing the prompt used for each classification.

        include_response (bool, default=False): If True, includes a column named `response`
            in the output dataframe containing the raw response from the LLM.

        max_retries (int, optional): The maximum number of times to retry on exceptions.
            Defaults to 10.

        exit_on_error (bool, default=True): If True, stops processing evals after all retries
            are exhausted on a single eval attempt. If False, all evals are attempted before
            returning, even if some fail.

        run_sync (bool, default=False): If True, forces synchronous request submission.
            Otherwise evaluations will be run asynchronously if possible.

        concurrency (Optional[int], default=None): The number of concurrent evals if async
            submission is possible. If not provided, a recommended default concurrency is
            set on a per-model basis.

        progress_bar_format(Optional[str]): An optional format for progress bar shown. If not
            specified, defaults to: llm_classify |{bar}| {n_fmt}/{total_fmt} ({percentage:3.1f}%) "
            "| ⏳ {elapsed}<{remaining} | {rate_fmt}{postfix}". If 'None' is passed in specifically,
            the progress_bar log will be disabled.

    Returns:
        pandas.DataFrame: A dataframe where the `label` column (at column position 0) contains
            the classification labels. If provide_explanation=True, then an additional column named
            `explanation` is added to contain the explanation for each label. The dataframe has
            the same length and index as the input dataframe. The classification label values are
            from the entries in the rails argument or "NOT_PARSABLE" if the model's output could
            not be parsed. The output dataframe also includes three additional columns in the
            output dataframe: `exceptions`, `execution_status`, and `execution_seconds` containing
            details about execution errors that may have occurred during the classification as well
            as the total runtime of each classification (in seconds).
    """
    concurrency = concurrency or model.default_concurrency
    # clients need to be reloaded to ensure that async evals work properly
    model.reload_client()

    use_openai_function_call = (
        use_function_calling_if_available
        and isinstance(model, OpenAIModel)
        and model.supports_function_calling
    )

    model_kwargs = (
        openai_function_call_kwargs(rails, provide_explanation) if use_openai_function_call else {}
    )

    eval_template = normalize_classification_template(rails=rails, template=template)

    prompt_options = PromptOptions(provide_explanation=provide_explanation)

    labels: Iterable[Optional[str]] = [None] * len(data)
    explanations: Iterable[Optional[str]] = [None] * len(data)

    printif(verbose, f"Using prompt:\n\n{eval_template.prompt(prompt_options)}")
    if generation_info := model.verbose_generation_info():
        printif(verbose, generation_info)

    def _map_template(data: pd.Series[Any]) -> MultimodalPrompt:
        try:
            variables = {var: data[var] for var in eval_template.variables}
            empty_keys = [k for k, v in variables.items() if v is None]
            if empty_keys:
                raise PhoenixTemplateMappingError(
                    f"Missing template variables: {', '.join(empty_keys)}"
                )
            return eval_template.format(
                variable_values=variables,
                options=prompt_options,
            )
        except KeyError as exc:
            raise PhoenixTemplateMappingError(f"Missing template variable: {exc}")

    def _process_response(response: str) -> Tuple[str, Optional[str]]:
        if not use_openai_function_call:
            if provide_explanation:
                unrailed_label, explanation = (
                    eval_template.extract_label_from_explanation(response),
                    response,
                )
                printif(
                    verbose and unrailed_label == NOT_PARSABLE,
                    f"- Could not parse {repr(response)} while extracting label and explanation",
                )
            else:
                unrailed_label = response
                explanation = None
        else:
            unrailed_label, explanation = parse_openai_function_call(response)
        return snap_to_rail(unrailed_label, rails, verbose=verbose), explanation

    def _normalize_to_series(
        data: PROCESSOR_TYPE,
    ) -> pd.Series[Any]:
        if isinstance(data, pd.Series):
            return data

        variable_count = len(eval_template.variables)
        if variable_count == 1:
            return pd.Series({eval_template.variables[0]: data})
        elif variable_count > 1:
            if isinstance(data, str):
                raise ValueError("The data cannot be mapped to the template variables")
            elif isinstance(data, Iterable):
                return pd.Series(
                    {
                        template_var: input_val
                        for template_var, input_val in zip(eval_template.variables, data)
                    }
                )
        return pd.Series()

    async def _run_llm_classification_async(
        input_data: PROCESSOR_TYPE,
    ) -> ParsedLLMResponse:
        with set_verbosity(model, verbose) as verbose_model:
            if data_processor:
                maybe_awaitable_data = data_processor(input_data)
                if inspect.isawaitable(maybe_awaitable_data):
                    processed_data = await maybe_awaitable_data
                else:
                    processed_data = maybe_awaitable_data
            else:
                processed_data = input_data

            prompt = _map_template(_normalize_to_series(processed_data))
            response = await verbose_model._async_generate(
                prompt, instruction=system_instruction, **model_kwargs
            )
        inference, explanation = _process_response(response)
        return inference, explanation, response, str(prompt)

    def _run_llm_classification_sync(
        input_data: PROCESSOR_TYPE,
    ) -> ParsedLLMResponse:
        with set_verbosity(model, verbose) as verbose_model:
            if data_processor:
                processed_data = data_processor(input_data)
                if inspect.isawaitable(processed_data):
                    raise ValueError("Cannot run the data processor asynchronously.")
            else:
                processed_data = input_data

            prompt = _map_template(_normalize_to_series(processed_data))
            response = verbose_model._generate(
                prompt, instruction=system_instruction, **model_kwargs
            )
        inference, explanation = _process_response(response)
        return inference, explanation, response, str(prompt)

    fallback_return_value: ParsedLLMResponse = (None, None, "", "")

    executor = get_executor_on_sync_context(
        _run_llm_classification_sync,
        _run_llm_classification_async,
        run_sync=run_sync,
        concurrency=concurrency,
        tqdm_bar_format=progress_bar_format,
        max_retries=max_retries,
        exit_on_error=exit_on_error,
        fallback_return_value=fallback_return_value,
        timeout=model._timeout,
    )

    list_of_inputs: Union[Tuple[Any], List[Any]]
    if isinstance(data, pd.DataFrame):
        list_of_inputs = [row_tuple[1] for row_tuple in data.iterrows()]
        dataframe_index = data.index
    elif isinstance(data, (list, tuple)):
        list_of_inputs = data
        dataframe_index = pd.Index(range(len(data)))
    else:
        raise ValueError("Invalid 'data' input type.")

    results, execution_details = executor.run(list_of_inputs)
    labels, explanations, responses, prompts = zip(*results)
    all_exceptions = [details.exceptions for details in execution_details]
    execution_statuses = [details.status for details in execution_details]
    execution_times = [details.execution_seconds for details in execution_details]
    classification_statuses = []
    for exceptions, status in zip(all_exceptions, execution_statuses):
        if exceptions and isinstance(exceptions[-1], PhoenixTemplateMappingError):
            classification_statuses.append(ClassificationStatus.MISSING_INPUT)
        else:
            classification_statuses.append(ClassificationStatus(status.value))

    return pd.DataFrame(
        data={
            "label": labels,
            **({"explanation": explanations} if provide_explanation else {}),
            **({"prompt": prompts} if include_prompt else {}),
            **({"response": responses} if include_response else {}),
            **({"exceptions": [[repr(exc) for exc in excs] for excs in all_exceptions]}),
            **({"execution_status": [status.value for status in classification_statuses]}),
            **({"execution_seconds": [runtime for runtime in execution_times]}),
        },
        index=dataframe_index,
    )


class RunEvalsPayload(NamedTuple):
    evaluator: LLMEvaluator
    record: Record


def run_evals(
    dataframe: DataFrame,
    evaluators: List[LLMEvaluator],
    provide_explanation: bool = False,
    use_function_calling_if_available: bool = True,
    verbose: bool = False,
    concurrency: Optional[int] = None,
) -> List[DataFrame]:
    """
    Applies a list of evaluators to a dataframe. Outputs a list of dataframes in
    which each dataframe contains the outputs of the corresponding evaluator
    applied to the input dataframe.

    Args:
        dataframe (DataFrame): A pandas dataframe in which each row represents a
            record to be evaluated. All template variable names must appear as
            column names in the dataframe (extra columns unrelated to the template
            are permitted).

        evaluators (List[LLMEvaluator]): A list of evaluators.

        provide_explanation (bool, optional): If True, provides an explanation
            for each evaluation. A column named "explanation" is added to each
            output dataframe.

        use_function_calling_if_available (bool, optional): If True, use
            function calling (if available) as a means to constrain the LLM outputs.
            With function calling, the LLM is instructed to provide its response as
            a structured JSON object, which is easier to parse.

        verbose (bool, optional): If True, prints detailed info to stdout such
            as model invocation parameters and details about retries and snapping to
            rails.

        concurrency (Optional[int], default=None): The number of concurrent evals
            if async submission is possible. If not provided, a recommended default
            concurrency is set on a per-model basis.

    Returns:
        List[DataFrame]: A list of dataframes, one for each evaluator, all of
            which have the same number of rows as the input dataframe.
    """
    # use the minimum default concurrency of all the models
    if concurrency is None:
        if len(evaluators) == 0:
            concurrency = 1
        else:
            concurrency = min(evaluator.default_concurrency for evaluator in evaluators)

    # use the maximum timeout of all the evaluators
    timeout = max(
        (
            evaluator._model._timeout
            for evaluator in evaluators
            if evaluator._model._timeout is not None
        ),
        default=None,
    )

    # clients need to be reloaded to ensure that async evals work properly
    for evaluator in evaluators:
        evaluator.reload_client()

    async def _arun_eval(
        payload: RunEvalsPayload,
    ) -> Tuple[Label, Score, Explanation]:
        return await payload.evaluator.aevaluate(
            payload.record,
            provide_explanation=provide_explanation,
            use_function_calling_if_available=use_function_calling_if_available,
            verbose=verbose,
        )

    def _run_eval(
        payload: RunEvalsPayload,
    ) -> Tuple[Label, Score, Explanation]:
        return payload.evaluator.evaluate(
            payload.record,
            provide_explanation=provide_explanation,
            use_function_calling_if_available=use_function_calling_if_available,
            verbose=verbose,
        )

    executor = get_executor_on_sync_context(
        _run_eval,
        _arun_eval,
        concurrency=concurrency,
        tqdm_bar_format=get_tqdm_progress_bar_formatter("run_evals"),
        exit_on_error=True,
        fallback_return_value=(None, None, None),
        timeout=timeout,
    )

    total_records = len(dataframe)
    payloads = [
        RunEvalsPayload(evaluator=evaluator, record=row)
        for evaluator, (_, row) in product(evaluators, dataframe.iterrows())
    ]
    eval_results: List[DefaultDict[Index, Dict[ColumnName, Union[Label, Explanation]]]] = [
        defaultdict(dict) for _ in range(len(evaluators))
    ]
    results, _ = executor.run(payloads)
    for index, (label, score, explanation) in enumerate(results):
        evaluator_index = index // total_records
        row_index = index % total_records
        eval_results[evaluator_index][row_index]["label"] = label
        eval_results[evaluator_index][row_index]["score"] = score
        if provide_explanation:
            eval_results[evaluator_index][row_index]["explanation"] = explanation
    eval_dataframes: List[DataFrame] = []
    for eval_result in eval_results:
        eval_data = [eval_result[row_index] for row_index in range(len(eval_result))]
        eval_dataframes.append(DataFrame(eval_data, index=dataframe.index))
    return eval_dataframes



import base64
import logging
import warnings
from dataclasses import dataclass, field, fields
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
    get_args,
    get_origin,
)
from urllib.parse import urlparse

from phoenix.evals.exceptions import PhoenixContextLimitExceeded, PhoenixUnsupportedAudioFormat
from phoenix.evals.models.base import BaseModel
from phoenix.evals.models.rate_limiters import RateLimiter
from phoenix.evals.templates import MultimodalPrompt, PromptPartContentType
from phoenix.evals.utils import get_audio_format_from_base64

MINIMUM_OPENAI_VERSION = "1.0.0"
MODEL_TOKEN_LIMIT_MAPPING = {
    "gpt-3.5-turbo-instruct": 4096,
    "gpt-3.5-turbo-0301": 4096,
    "gpt-3.5-turbo-0613": 4096,  # Current gpt-3.5-turbo default
    "gpt-3.5-turbo-16k-0613": 16385,
    "gpt-4-0314": 8192,
    "gpt-4-0613": 8192,  # Current gpt-4 default
    "gpt-4-32k-0314": 32768,
    "gpt-4-32k-0613": 32768,
    "gpt-4-1106-preview": 128000,
    "gpt-4-0125-preview": 128000,
    "gpt-4-turbo-preview": 128000,
    "gpt-4-vision-preview": 128000,
}
LEGACY_COMPLETION_API_MODELS = ("gpt-3.5-turbo-instruct",)
SUPPORTED_AUDIO_FORMATS = {"mp3", "wav"}
logger = logging.getLogger(__name__)


@dataclass
class AzureOptions:
    api_version: str
    azure_endpoint: str
    azure_deployment: Optional[str]
    azure_ad_token: Optional[str]
    azure_ad_token_provider: Optional[Callable[[], str]]


def _model_supports_temperature(model: str) -> bool:
    """OpenAI reasoning models do not support temperature."""
    if model.startswith("o1") or model.startswith("o3"):
        return False
    return True


@dataclass
class OpenAIModel(BaseModel):
    """
    An interface for using OpenAI models.

    This class wraps the OpenAI SDK library for use with Phoenix LLM evaluations. Calls to the
    OpenAI API are dynamically throttled when encountering rate limit errors. Requires the
    `openai` package to be installed.

    Additionally, OpenAIModel supports Azure OpenAI API. To use Azure OpenAI API, you need to
    provide the `azure_endpoint` and `azure_deployment` parameters. You can also provide the
    `azure_ad_token` or `azure_ad_token_provider` to authenticate with Azure OpenAI API.

    Supports Async: ✅
        If possible, makes LLM calls concurrently.

    Args:
        api_key (str, optional): Your OpenAI key. If not provided, will be read from the
            environment variable. Defaults to None.
        organization (str, optional): The organization to use for the OpenAI API. If not provided,
            will default to what's configured in OpenAI. Defaults to None.
        base_url (str, optional): An optional base URL to use for the OpenAI API. If not provided,
            will default to what's configured in OpenAI. Defaults to None.
        model (str, optional): Model name to use. In of azure, this is the deployment name such as
            gpt-35-instant. Defaults to "gpt-4".
        temperature (float, optional): What sampling temperature to use. Defaults to 0.0.
        max_tokens (int | None, optional): The maximum number of tokens to generate in the
            completion. To unset this limit, set `max_tokens` to `None`. Defaults to 256.
        top_p (float, optional): Total probability mass of tokens to consider at each step.
            Defaults to 1.
        frequency_penalty (float, optional): Penalizes repeated tokens according to frequency.
            Defaults to 0.
        presence_penalty (float, optional): Penalizes repeated tokens. Defaults to 0.
        n (int, optional): How many completions to generate for each prompt. Defaults to 1.
        model_kwargs (Dict[str, Any], optional): Holds any model parameters valid for `create` call
            not explicitly specified. Defaults to an empty dictionary.
        request_timeout (Optional[Union[float, Tuple[float, float]]], optional): Timeout for
            requests to OpenAI completion API. Default is 600 seconds. Defaults to None.
        initial_rate_limit (int, optional): The initial internal rate limit in allowed requests
            per second for making LLM calls. This limit adjusts dynamically based on rate
            limit errors. Defaults to 10.
        timeout (int, optional): Pheonix timeout for API requests in seconds. Defaults to 120.
        api_version (str, optional): The version of the Azure API to use. Defaults to None.
            https://learn.microsoft.com/en-us/azure/ai-services/openai/reference#rest-api-versioning
        azure_endpoint (str, optional): The endpoint to use for azure openai. Available in the
            Azure portal. Defaults to None.
        azure_deployment (str, optional): The deployment to use for azure openai. Defaults to None.
        azure_ad_token (str, optional): The azure AD token to use for azure openai. Defaults to
            None.
        azure_ad_token_provider (Callable[[], str], optional): A callable that returns the azure ad
            token to use for azure openai. Defaults to None.
        default_headers (Mapping[str, str], optional): Default headers required by AzureOpenAI.
            Defaults to None.

    Examples:
        After setting the OPENAI_API_KEY environment variable:
        .. code-block:: python

            from phoenix.evals import OpenAIModel
            model = OpenAIModel(model="gpt-4o")

        Using OpenAI models via Azure is similar (after setting the AZURE_OPENAI_API_KEY
        environment variable):

        .. code-block:: python

            from phoenix.evals import OpenAIModel
            model = OpenAIModel(
                model="gpt-35-turbo-16k",
                azure_endpoint="https://your-endpoint.azure.com/",
                api_version="2023-09-15-preview",
            )
    """

    api_key: Optional[str] = field(repr=False, default=None)
    organization: Optional[str] = field(repr=False, default=None)
    base_url: Optional[str] = field(repr=False, default=None)
    model: str = "gpt-4"
    temperature: float = 0.0
    max_tokens: Optional[int] = None
    top_p: float = 1
    frequency_penalty: float = 0
    presence_penalty: float = 0
    n: int = 1
    model_kwargs: Dict[str, Any] = field(default_factory=dict)
    request_timeout: Optional[Union[float, Tuple[float, float]]] = None
    initial_rate_limit: int = 10
    timeout: int = 120

    # Azure options
    api_version: Optional[str] = field(default=None)
    azure_endpoint: Optional[str] = field(default=None)
    azure_deployment: Optional[str] = field(default=None)
    azure_ad_token: Optional[str] = field(default=None)
    azure_ad_token_provider: Optional[Callable[[], str]] = field(default=None)
    default_headers: Optional[Mapping[str, str]] = field(default=None)

    # Deprecated fields
    model_name: Optional[str] = field(default=None)
    """
    .. deprecated:: 3.0.0
       use `model` instead. This will be removed
    """

    def __post_init__(self) -> None:
        self._migrate_model_name()
        self._init_environment()
        self._init_open_ai()
        self._init_rate_limiter()

    @property
    def _model_name(self) -> str:
        return self.model

    def reload_client(self) -> None:
        self._init_open_ai()

    def _migrate_model_name(self) -> None:
        if self.model_name:
            warning_message = "The `model_name` field is deprecated. Use `model` instead. \
                This will be removed in a future release."
            print(
                warning_message,
            )
            warnings.warn(warning_message, DeprecationWarning)
            self.model = self.model_name
            self.model_name = None

    def _init_environment(self) -> None:
        try:
            import openai
            import openai._utils as openai_util

            self._openai = openai
            self._openai_util = openai_util
        except ImportError:
            self._raise_import_error(
                package_display_name="OpenAI",
                package_name="openai",
                package_min_version=MINIMUM_OPENAI_VERSION,
            )

    def _init_open_ai(self) -> None:
        # For Azure, you need to provide the endpoint and the endpoint
        self._is_azure = bool(self.azure_endpoint)

        self._model_uses_legacy_completion_api = self.model.startswith(LEGACY_COMPLETION_API_MODELS)

        # Set the version, organization, and base_url - default to openAI
        self.api_version = self.api_version or self._openai.api_version
        self.organization = self.organization or self._openai.organization

        # Initialize specific clients depending on the API backend
        # Set the type first
        self._client: Union[self._openai.OpenAI, self._openai.AzureOpenAI]  # type: ignore
        self._async_client: Union[self._openai.AsyncOpenAI, self._openai.AsyncAzureOpenAI]  # type: ignore
        if self._is_azure:
            # Validate the azure options and construct a client
            azure_options = self._get_azure_options()
            self._client = self._openai.AzureOpenAI(
                azure_endpoint=azure_options.azure_endpoint,
                azure_deployment=azure_options.azure_deployment,
                api_version=azure_options.api_version,
                azure_ad_token=azure_options.azure_ad_token,
                azure_ad_token_provider=azure_options.azure_ad_token_provider,
                api_key=self.api_key,
                organization=self.organization,
                default_headers=self.default_headers,
            )
            self._async_client = self._openai.AsyncAzureOpenAI(
                azure_endpoint=azure_options.azure_endpoint,
                azure_deployment=azure_options.azure_deployment,
                api_version=azure_options.api_version,
                azure_ad_token=azure_options.azure_ad_token,
                azure_ad_token_provider=azure_options.azure_ad_token_provider,
                api_key=self.api_key,
                organization=self.organization,
                default_headers=self.default_headers,
            )
            # return early since we don't need to check the model
            return

        # The client is not azure, so it must be openai
        self._client = self._openai.OpenAI(
            api_key=self.api_key,
            organization=self.organization,
            base_url=(self.base_url or self._openai.base_url),
        )

        # The client is not azure, so it must be openai
        self._async_client = self._openai.AsyncOpenAI(
            api_key=self.api_key,
            organization=self.organization,
            base_url=(self.base_url or self._openai.base_url),
            max_retries=0,
        )

    def _get_azure_options(self) -> AzureOptions:
        options = {}
        for option in fields(AzureOptions):
            if (value := getattr(self, option.name)) is not None:
                options[option.name] = value
            else:
                # raise ValueError if field is not optional
                # See if the field is optional - e.g. get_origin(Optional[T])  = typing.Union
                option_is_optional = get_origin(option.type) is Union and type(None) in get_args(
                    option.type
                )
                if not option_is_optional:
                    raise ValueError(
                        f"Option '{option.name}' must be set when using Azure OpenAI API"
                    )
                options[option.name] = None
        return AzureOptions(**options)

    def _init_rate_limiter(self) -> None:
        self._rate_limiter = RateLimiter(
            rate_limit_error=self._openai.RateLimitError,
            max_rate_limit_retries=10,
            initial_per_second_request_rate=self.initial_rate_limit,
            enforcement_window_minutes=1,
        )

    def _build_messages(
        self, prompt: MultimodalPrompt, system_instruction: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        system_role = self._system_role()
        messages: List[Dict[str, Any]] = []
        for part in prompt.parts:
            if part.content_type == PromptPartContentType.TEXT:
                messages.append({"role": system_role, "content": part.content})
            elif part.content_type == PromptPartContentType.AUDIO:
                format = str(get_audio_format_from_base64(part.content))
                if format not in SUPPORTED_AUDIO_FORMATS:
                    raise PhoenixUnsupportedAudioFormat(f"Unsupported audio format: {format}")
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "input_audio",
                                "input_audio": {
                                    "data": part.content,
                                    "format": str(get_audio_format_from_base64(part.content)),
                                },
                            }
                        ],
                    }
                )
            elif part.content_type == PromptPartContentType.IMAGE:
                if _is_base64(part.content):
                    content_url = f"data:image/jpeg;base64,{part.content}"
                elif _is_url(part.content):
                    content_url = part.content
                else:
                    raise ValueError("Only base64 encoded images or image URLs are supported")
                messages.append(
                    {
                        "role": "user",
                        "content": [
                            {
                                "type": "image_url",
                                "image_url": {"url": content_url},
                            }
                        ],
                    }
                )
            else:
                raise ValueError(
                    f"Unsupported content type for {OpenAIModel.__name__}: {part.content_type}"
                )
        if system_instruction:
            messages.insert(0, {"role": system_role, "content": str(system_instruction)})
        return messages

    def verbose_generation_info(self) -> str:
        return f"OpenAI invocation parameters: {self.public_invocation_params}"

    async def _async_generate(self, prompt: Union[str, MultimodalPrompt], **kwargs: Any) -> str:
        if isinstance(prompt, str):
            prompt = MultimodalPrompt.from_string(prompt)

        invoke_params = self.invocation_params
        messages = self._build_messages(prompt, kwargs.get("instruction"))
        if functions := kwargs.get("functions"):
            invoke_params["functions"] = functions
        if function_call := kwargs.get("function_call"):
            invoke_params["function_call"] = function_call
        response = await self._async_rate_limited_completion(
            messages=messages,
            **invoke_params,
        )
        choice = response["choices"][0]
        if self._model_uses_legacy_completion_api:
            return str(choice["text"])
        message = choice["message"]
        if function_call := message.get("function_call"):
            return str(function_call.get("arguments") or "")
        if tool_calls := message.get("tool_calls"):
            try:
                for tool_call in tool_calls:
                    if tool_call.get("type") == "function":
                        return str((tool_call.get("function") or {}).get("arguments") or "")
            except Exception as e:
                logger.error(f"Error getting tool call arguments: {e}")
        return str(message["content"])

    def _generate(self, prompt: Union[str, MultimodalPrompt], **kwargs: Any) -> str:
        if isinstance(prompt, str):
            prompt = MultimodalPrompt.from_string(prompt)

        invoke_params = self.invocation_params
        messages = self._build_messages(prompt=prompt, system_instruction=kwargs.get("instruction"))
        if functions := kwargs.get("functions"):
            invoke_params["functions"] = functions
        if function_call := kwargs.get("function_call"):
            invoke_params["function_call"] = function_call
        response = self._rate_limited_completion(
            messages=messages,
            **invoke_params,
        )
        choice = response["choices"][0]
        if self._model_uses_legacy_completion_api:
            return str(choice["text"])
        message = choice["message"]
        if function_call := message.get("function_call"):
            return str(function_call.get("arguments") or "")
        if tool_calls := message.get("tool_calls"):
            try:
                for tool_call in tool_calls:
                    if tool_call.get("type") == "function":
                        return str((tool_call.get("function") or {}).get("arguments") or "")
            except Exception as e:
                logger.error(f"Error getting tool call arguments: {e}")
        return str(message["content"])

    async def _async_rate_limited_completion(self, **kwargs: Any) -> Any:
        @self._rate_limiter.alimit
        async def _async_completion(**kwargs: Any) -> Any:
            try:
                if self._model_uses_legacy_completion_api:
                    if "prompt" not in kwargs:
                        kwargs["prompt"] = "\n\n".join(
                            (message.get("content") or "")
                            for message in (kwargs.pop("messages", None) or ())
                        )
                    # OpenAI 1.0.0 API responses are pydantic objects, not dicts
                    # We must dump the model to get the dict
                    res = await self._async_client.completions.create(**kwargs)
                else:
                    res = await self._async_client.chat.completions.create(**kwargs)
                return res.model_dump()
            except self._openai._exceptions.BadRequestError as e:
                exception_message = e.args[0]
                if exception_message and "maximum context length" in exception_message:
                    raise PhoenixContextLimitExceeded(exception_message) from e
                raise e

        return await _async_completion(**kwargs)

    def _rate_limited_completion(self, **kwargs: Any) -> Any:
        @self._rate_limiter.limit
        def _completion(**kwargs: Any) -> Any:
            try:
                if self._model_uses_legacy_completion_api:
                    if "prompt" not in kwargs:
                        kwargs["prompt"] = "\n\n".join(
                            (message.get("content") or "")
                            for message in (kwargs.pop("messages", None) or ())
                        )
                    # OpenAI 1.0.0 API responses are pydantic objects, not dicts
                    # We must dump the model to get the dict
                    return self._client.completions.create(**kwargs).model_dump()
                return self._client.chat.completions.create(**kwargs).model_dump()
            except self._openai._exceptions.BadRequestError as e:
                exception_message = e.args[0]
                if exception_message and "maximum context length" in exception_message:
                    raise PhoenixContextLimitExceeded(exception_message) from e
                raise e

        return _completion(**kwargs)

    def _system_role(self) -> str:
        # OpenAI uses different semantics for "system" roles for different models
        if "gpt" in self.model:
            return "system"
        if "o1-mini" in self.model:
            return "user"  # o1-mini does not support either "system" or "developer" roles
        if "o1-preview" in self.model:
            return "user"  # o1-preview does not support "system" or "developer" roles
        if "o1" in self.model:
            return "developer"
        if "o3" in self.model:
            return "developer"
        return "system"

    @property
    def public_invocation_params(self) -> Dict[str, Any]:
        return {
            **({"model": self.model}),
            **self._default_params,
            **self.model_kwargs,
        }

    @property
    def invocation_params(self) -> Dict[str, Any]:
        return {
            **self.public_invocation_params,
        }

    @property
    def _default_params(self) -> Dict[str, Any]:
        """Get the default parameters for calling OpenAI API."""
        # token param str depends on provider and model
        token_param_str = _get_token_param_str(self._is_azure, self.model)
        params = {
            "frequency_penalty": self.frequency_penalty,
            "presence_penalty": self.presence_penalty,
            "top_p": self.top_p,
            "n": self.n,
            "timeout": self.request_timeout,
            token_param_str: self.max_tokens,
        }
        if _model_supports_temperature(self.model):
            params.update(
                {
                    "temperature": self.temperature,
                }
            )
        return params

    @property
    def supports_function_calling(self) -> bool:
        if (
            self._is_azure
            and self.api_version
            # The first api version supporting function calling is 2023-07-01-preview.
            # See https://github.com/Azure/azure-rest-api-specs/blob/58e92dd03733bc175e6a9540f4bc53703b57fcc9/specification/cognitiveservices/data-plane/AzureOpenAI/inference/preview/2023-07-01-preview/inference.json#L895 # noqa E501
            and self.api_version[:10] < "2023-07-01"
        ):
            return False
        if self._model_uses_legacy_completion_api:
            return False
        if self.model.startswith("o1"):
            return False
        return True


def _is_url(url: str) -> bool:
    parsed_url = urlparse(url)
    return bool(parsed_url.scheme and parsed_url.netloc)


def _is_base64(s: str) -> bool:
    try:
        base64.b64decode(s, validate=True)
        return True
    except Exception:
        return False


def _get_token_param_str(is_azure: bool, model: str) -> str:
    """
    Get the token parameter string for the given model.
    OpenAI o1 and o3 models made a switch to use
    max_completion_tokens and now all the models support it.
    However, Azure OpenAI models currently do not support
    max_completion_tokens unless it's an o1 or o3 model.
    """
    if is_azure and not model.startswith("o1") and not model.startswith("o3"):
        return "max_tokens"
    return "max_completion_tokens"
